@@ -1,18 +1,20 @@
 import ast
+from typing import List
 
 import click
 import os
 import shutil
 import re
-import sys
 import platform
 import subprocess
 import site
 import webbrowser
+import hashlib
 from datetime import datetime
 from pathlib import Path
-from json import load as json_load
-from json import dumps as json_dumps
+from ujson import load as json_load
+from ujson import dumps as json_dumps
+from ujson import dump as json_dump
 from dsframework.documentation.scripts import doxygen_handler
 
 isWindows = platform.system() == 'Windows'
@@ -20,6 +22,8 @@ isWindows = platform.system() == 'Windows'
 __location__ = os.path.realpath(
     os.path.join(os.getcwd(), os.path.dirname(__file__)))
 
+__proj_root__ = os.path.abspath(os.getcwd())
+
 directories = {
     'main': '',
     'pipeline': '',
@@ -41,6 +45,7 @@ directories = {
     'documentation': ''
 }
 
+
 ##
 # @file
 # @brief This file defines and implements the framework CLI. It helps us to:\n
@@ -83,14 +88,14 @@ def cli():
     ====================
     dsf-cli g predictable my-new-predictable
 
-    Generate stage
-    ====================
-    dsf-cli g stage my-new-stage
-
     Create files
     ===================
     dsf-cli create-files [tester / deploy / cloud_eval / documentation / batch / trainer]
 
+    Create Mlops dir
+    ====================
+    dsf-cli create-mlops
+
     run server
     ==========
     dsf-cli run-server
@@ -123,10 +128,18 @@ def cli():
     Batch project abilities
     ========================
 
+    Workflow Creation
+    =================
+    dsf-cli generate workflow <new_workflow_name>
+
     DataProc Cluster Manipulation
     =============================
     dsf-cli dataproc-cluster <create / stop / start / delete / connect>
 
+    Batch Stage Manipulation
+    ========================
+    dsf-cli batch-stage <create / run> my-stage
+
     Workflow Template Manipulation
     ==============================
     dsf-cli workflow-template <create / create_yaml / delete / instantiate>
@@ -156,6 +169,8 @@ def cli():
     =====================
     dsf-cli deploy_service_to_gae
     """
+
+
 # @click.option("--type", prompt="type", help="type of component")
 # @click.option("--project_name", prompt="project_name", help="project_name")
 # def apis(type, project_name):
@@ -194,11 +209,19 @@ def create_files(file_type):
     f()
 
 
+@cli.command()
+def create_mlops():
+    """! Copy Mlops directory from DSF """
+    source_dir = os.path.join(__location__, 'dsframework/cli/mlops/')
+    target_dir = os.path.join(os.path.abspath(os.getcwd()), 'mlops')
+    shutil.copytree(source_dir, target_dir)
+
+
 @cli.command()
 def run_server():
     """! Start a uvicorn server by running /server/main.py """
-    currentPipelineFolder = os.path.abspath(os.getcwd())
-    currentParentFolder = os.path.join(currentPipelineFolder,"..")
+    currentPipelineFolder = __proj_root__
+    currentParentFolder = os.path.join(currentPipelineFolder, "..")
     os.environ["PYTHONPATH"] = currentPipelineFolder
     folder = currentPipelineFolder + '/server/main.py'
     subprocess.call('python ' + folder, shell=True)
@@ -209,8 +232,8 @@ def run_server():
 @click.argument('env', default="prd")
 def cloud_eval(tag, env):
     # """Cloud evaluation running on AWS """
-    currentPipelineFolder = os.path.abspath(os.getcwd())
-    currentParentFolder = os.path.join(currentPipelineFolder,"..")
+    currentPipelineFolder = __proj_root__
+    currentParentFolder = os.path.join(currentPipelineFolder, "..")
     os.environ["PYTHONPATH"] = currentPipelineFolder
     folder = currentPipelineFolder + '/tester/dsp/get_model_options.py'
     output = subprocess.check_output('python ' + folder + ' ' + env, shell=True, encoding='utf-8').strip()
@@ -227,8 +250,9 @@ def cloud_eval(tag, env):
                 'repo': o['repo'],
             }
             folder = currentPipelineFolder + '/tester/dsp/get_dataset_options.py'
-            output = subprocess.check_output('python ' + folder + ' ' + env + ' ' + model['name'] + ' ' + str(model['id']),
-                                             shell=True, encoding='utf-8').strip()
+            output = subprocess.check_output(
+                'python ' + folder + ' ' + env + ' ' + model['name'] + ' ' + str(model['id']),
+                shell=True, encoding='utf-8').strip()
             l = output.split('\n')
             m = l[-1]
             l = ast.literal_eval(m)
@@ -240,7 +264,8 @@ def cloud_eval(tag, env):
                 # print('selected_dataset', selected_dataset)
                 folder = currentPipelineFolder + '/tester/dsp/add_experiment.py'
                 subprocess.call(('dvc_push.sh'), shell=True)
-                output = subprocess.call(('python', folder, env, json_dumps(model), json_dumps(selected_dataset), tag), shell=True)
+                output = subprocess.call(('python', folder, env, json_dumps(model), json_dumps(selected_dataset), tag),
+                                         shell=True)
                 print('you can see your experiment in que')
             else:
                 print('selection does not exist')
@@ -257,8 +282,7 @@ def create_deploy_files():
 
 def create_cloud_eval_files():
     """! Create cloud evaluation files (created through 'Create project', but can be used as a standalone function) """
-    currentPipelineFolder = os.path.abspath(os.getcwd())
-    copy_cloud_eval_files(currentPipelineFolder)
+    copy_cloud_eval_files(__proj_root__)
 
 
 def create_tester_files():
@@ -284,18 +308,24 @@ def create_trainer_files():
 
 
 @cli.command()
+@click.argument('wf_name')
 @click.argument('action')
-def dataproc_cluster(action: str):
+def dataproc_cluster(wf_name: str, action: str):
     """! Handle cluster action, either create,  start, stop, delete, or connect to the Dataproc cluster
             Args:
-                 action: an action to perform, chosen from: {create, start, stop, delete, connect}
+                wf_name: Relevant workflow that affects cluster configuration
+                action: an action to perform, chosen from: {create, start, stop, delete, connect}
     """
     valid_actions = {'create', 'start', 'connect', 'stop', 'delete'}
     if action not in valid_actions:
-        print(f"Can't determine file type {action}. Please use one of: {valid_actions}")
+        print(f"Can't determine action of type: {action}. Please use one of: {valid_actions}")
         return
 
-    batch_helper = config_batch_helper()
+    if not workflow_exists(wf_name):
+        click.echo(f"Chosen workflow ({wf_name}) does not exist. Please choose a different workflow")
+        return
+
+    batch_helper = config_batch_helper(wf_name)
 
     if action == 'create':
         # Create DataProc cluster according to project specific configuration
@@ -317,73 +347,180 @@ def dataproc_cluster(action: str):
         batch_helper.delete_dataproc_cluster()
 
 
+def update_config_file(stage_name: str, workflow_name: str, prerequisites: List[str]):
+    """! Update relevant workflow configuration file with new stage
+    @verbatim
+    Args:
+            stage_name: Stage name to be used
+            workflow_name: Relevant workflow that will handle the new stage
+            prerequisites: A list of stages that the new stage depends on
+    @endverbatim
+    """
+
+    config_path = os.path.join(__proj_root__,
+                               f'batch_files/workflows/{workflow_name}_workflow/config_{workflow_name}.json')
+
+    stage_dict_entry = {"stage_name": stage_name,
+                        "prerequisites": prerequisites}
+
+    # Open json file, load to dictionary, update it and save it back
+    with open(config_path, 'r+') as config_file:
+        data = json_load(config_file)
+        data['workflow_stages'].append(stage_dict_entry)
+        config_file.seek(0)  # <--- should reset file position to the beginning.
+        json_dump(data, config_file, indent=4)
+        config_file.truncate()  # remove remaining part
+
+
+def create_batch_stage(stage_name: str = '', workflow_name: str = ""):
+    """! Generating new stage for a given workflow
+    @verbatim
+    Args:
+            stage_name: Stage name to be used when creating files / folders / classes.
+            workflow_name: Relevant workflow that will handle the new stage
+    @endverbatim
+    """
+    stage_name = clean_name(stage_name)
+    workflow_name = clean_name(workflow_name)
+
+    if not validate_stage(stage_name, workflow_name):
+        return
+
+    create_stage_files_from_template(stage_name, workflow_name)
+
+    # TBD: Should the prerequisites be validated, or this is a user responsibility?
+    prerequisites = click.prompt('Please enter dependency stages (separate with comma)',
+                                 type=str,
+                                 default="",
+                                 show_default=False)
+
+    prerequisite_list = [] if prerequisites == "" else [list_value.strip() for list_value in prerequisites.split(',')]
+
+    update_config_file(stage_name, workflow_name, prerequisite_list)
+
+    click.echo(f'Created {stage_name} stage under {workflow_name} Workflow')
+
+
 @cli.command()
 @click.argument('action')
-def workflow_template(action: str):
-    """! Handle workflow action, either create, instantiate or delete the template
+@click.argument('stage_name')
+def batch_stage(action: str, stage_name: str):
+    """! Handle batch stages action, either create or run a specific stage
         Args:
              action: an action to perform, chosen from: {create, create_yaml, instantiate, delete}
+             stage_name: Name for the newly created stage, or the stage to be run on an existing cluster
+    """
+    valid_actions = {'create', 'run'}
+    if action not in valid_actions:
+        print(f"Can't determine action of type: {action}. Please use one of: {valid_actions}")
+        return
+
+    relevant_workflow = click.prompt('Please enter a valid workflow name', type=str)
+
+    if action == 'create':
+        # Create a batch stage
+        create_batch_stage(stage_name, relevant_workflow)
+    elif action == 'run':
+        # Run specific stage on an existing Dataproc cluster
+        batch_helper = config_batch_helper(relevant_workflow)
+        batch_helper.run_single_stage(stage_name)
+
+
+@cli.command()
+@click.argument('wf_name')
+@click.argument('action')
+def workflow_template(wf_name: str, action: str):
+    """! Handle workflow action, either create, instantiate or delete the template
+        Args:
+            wf_name: Relevant workflow to perform the selected action
+            action: An action to perform, chosen from: {create, create_yaml, instantiate, delete}
     """
     valid_actions = {'create', 'create_yaml', 'delete', 'instantiate'}
     if action not in valid_actions:
-        print(f"Can't determine action type: {action}. Please use one of: {valid_actions}")
+        click.echo(f"Can't determine action type: {action}. Please use one of: {valid_actions}")
         return
 
-    current_folder = os.path.abspath(os.getcwd())
+    if not workflow_exists(wf_name):
+        click.echo(f"Chosen workflow ({wf_name}) does not exist. Please choose a different workflow")
+        return
+
+    current_folder = __proj_root__
 
     os.environ["PYTHONPATH"] = current_folder
-    workflow_folder = current_folder + '/batch_files/workflow/workflow.py'
+    workflow_folder = current_folder + '/batch_files/workflows/workflow.py'
     basic_command = 'python ' + workflow_folder
 
-    print(f"Attempting to {action} a workflow template")
-    subprocess.call(basic_command + ' ' + action, shell=True)
+    click.echo(f"Attempting to {action} a workflow template")
+    subprocess.call(basic_command + ' --wf-name ' + wf_name + ' --action ' + action, shell=True)
 
 
 @cli.command()
+@click.argument('wf_name')
 @click.argument('action')
-def dag(action: str):
+def dag(wf_name: str, action: str):
     """! Handle DAG actions, either create, delete, or import a DAG
     Once imported, it will start as paused and will be set up for a daily activation
         Args:
-             action: an action to perform, chosen from: {create, import, delete}
+            wf_name: Relevant workflow to perform the selected action
+            action: an action to perform, chosen from: {create, import, delete}
     """
     valid_actions = {'create', 'import', 'delete'}
     if action not in valid_actions:
-        print(f"Can't determine action type: {action}. Please use one of: {valid_actions}")
+        click.echo(f"Can't determine action type: {action}. Please use one of: {valid_actions}")
         return
 
-    current_folder = os.path.abspath(os.getcwd())
+    if not workflow_exists(wf_name):
+        click.echo(f"Chosen workflow ({wf_name}) does not exist. Please choose a different workflow")
+        return
+
+    current_folder = __proj_root__
 
     os.environ["PYTHONPATH"] = current_folder
-    dag_folder = current_folder + '/batch_files/workflow_dag/dag.py'
+    dag_folder = current_folder + '/batch_files/workflows/dag.py'
     basic_command = 'python ' + dag_folder
 
-    print(f"Attempting to {action} a DAG")
-    subprocess.call(basic_command + ' ' + action, shell=True)
+    click.echo(f"Attempting to {action} a DAG")
+    subprocess.call(basic_command + ' --wf-name ' + wf_name + ' --action ' + action, shell=True)
 
 
 @cli.command()
+@click.argument('wf_name')
 @click.argument('file_type')
-def upload_batch_files(file_type: str):
+def upload_batch_files(wf_name: str, file_type: str):
     """! Upload all files needed for a batch project to the predefined bucket
         Args:
+            wf_name: Relevant workflow to perform the selected action
             file_type: String containing the types of files to upload (jars, stages, whl , cluster_init, or "all")
     """
-    batch_helper = config_batch_helper()
+    if not workflow_exists(wf_name):
+        click.echo(f"Chosen workflow ({wf_name}) does not exist. Please choose a different workflow")
+        return
+
+    batch_helper = config_batch_helper(wf_name)
     batch_helper.upload_files_to_bucket(file_type)
 
 
-def config_batch_helper():
+def config_batch_helper(wf_name: str):
     """! Configure the batch file helper for a variety of cli actions
+        Args:
+            wf_name: Relevant workflow to perform the selected action
         Returns:
             ZIDSBatchHelper: Instantiation of the batch helper class
     """
     from dsframework.cli.batch.batch_helper_functions import ZIDSBatchHelper
-    current_folder = os.path.abspath(os.getcwd())
+
     batch_helper_cls = None
-    with open(current_folder + '/batch_files/batch_config.json') as config_file:
-        config_dict = json_load(config_file)
-        batch_helper_cls = ZIDSBatchHelper(config_dict)
+
+    with open(__proj_root__ + '/batch_files/batch_config.json') as batch_cfg_file:
+        batch_config = json_load(batch_cfg_file)
+
+    with open(__proj_root__ + f"/batch_files/workflows/{wf_name}_workflow/config_{wf_name}.json") as wf_cfg_file:
+        wf_config = json_load(wf_cfg_file)
+
+    config_dict = {**batch_config, **wf_config}
+
+    batch_helper_cls = ZIDSBatchHelper(config_dict)
+
     return batch_helper_cls
 
 
@@ -395,9 +532,9 @@ def create_sonar_file(repo_name: str):
     Args:
         repo_name: name of your repository.
     """
-    current_folder = os.path.abspath(os.getcwd())
+
     data = "sonar.projectKey=dozi_" + repo_name
-    create_file(os.path.join(current_folder, "sonar-scanner.properties"), data)
+    create_file(os.path.join(__proj_root__, "sonar-scanner.properties"), data)
 
 
 @cli.command()
@@ -414,7 +551,7 @@ def evaluate(csv_file_path, batch_size):
     """
 
     if os.path.isfile(csv_file_path):
-        current_pipeline_folder = os.path.abspath(os.getcwd())
+        current_pipeline_folder = __proj_root__
         os.environ["PYTHONPATH"] = current_pipeline_folder
         folder = current_pipeline_folder + '/tester/general_tester.py'
         subprocess.call('python ' + folder + ' ' + csv_file_path + ' ' + batch_size, shell=True)
@@ -429,7 +566,7 @@ def show_documentation():
 
 
 def launch_doc_webbrowser():
-    current_pipeline_folder = os.path.abspath(os.getcwd())
+    current_pipeline_folder = __proj_root__
 
     current_doc_folder = os.path.join(current_pipeline_folder, "documentation")
     if os.path.isdir(os.path.join(current_pipeline_folder, "dsframework", "documentation")):
@@ -475,29 +612,104 @@ def generate_predictable(file_name):
     create_exist_pipeline_file('predictable', file_name)
 
 
-def stage_name_already_in_use(stage_name):
-    """! Checks if a stage with the given name already exists
-    """
-    if os.path.exists('batch_files/stages/' + stage_name):
-        click.echo('Stage \'' + stage_name + '\' already exists. Please use a different name.')
-        return True
-    return False
-
-
-def generate_stage(stage_name):
-    """! Generating new stage (with the relevant project_name)\n
-    How to generate stage :
-    > dsf-cli g stage my-new-stage
+def add_workflow_to_setup(workflow_name: str):
+    """! Modify setup.py for the whl creation of batch projects
     @verbatim
     Args:
-            stage_name: Stage name to be used when creating files / folders / classes.
+            workflow_name: Workflow to be added to the file
     @endverbatim
     """
-    stage_name = clean_name(stage_name)
-    if stage_name_already_in_use(stage_name):
+
+    insert_string = f"if build_type == '{workflow_name}':\n\tdata = config + pipeline + batch_files + tester\n\n"
+    setup_file_path = os.path.join(__proj_root__, "batch_files/setup.py")
+
+    with open(setup_file_path, 'r+') as setup_fd:
+        contents = setup_fd.readlines()
+        if "# Additional workflows" in contents[-1]:  # Handle last line to prevent IndexError
+            contents.append(insert_string)
+        else:
+            for index, line in enumerate(contents):
+                if "# Additional workflows" in line and insert_string not in contents[index + 1]:
+                    contents.insert(index + 1, insert_string)
+                    break
+        setup_fd.seek(0)
+        setup_fd.writelines(contents)
+
+
+def generate_workflow(workflow_name: str):
+    """! Generating a new workflow to be used on GCP Dataproc
+    @verbatim
+    Args:
+            workflow_name: Workflow to be created
+    @endverbatim
+    """
+
+    click.echo('Generating workflow: ' + workflow_name)
+
+    if not os.path.exists(os.path.join(__proj_root__, "batch_files")):
+        click.echo("Couldn't find batch project folder, creating...")
+        copy_batch_files(directories['main'])
+
+    if workflow_exists(workflow_name):
+        click.echo(f'Workflow "{workflow_name}" already exists. Please use a different name.')
         return
-    create_stage_files_from_template(stage_name)
-    click.echo('Stage created: ' + stage_name)
+
+    workflow_path = os.path.join(__proj_root__, f"batch_files/workflows/{workflow_name}_workflow")
+
+    # Create primary workflow folder
+    os.mkdir(workflow_path)
+    create_empty_init_file(workflow_path)
+
+    # Create an empty stages folder
+    stages_path = os.path.join(workflow_path, "stages")
+    os.mkdir(stages_path)
+    create_empty_init_file(stages_path)
+
+    # Create an empty requirement file
+    create_file(os.path.join(workflow_path, f"requirements_{workflow_name}.txt"), "")
+
+    # Create configuration file from template
+    data = read_template_file('batch/workflow_config', "json")
+    data = data.replace('{name-your-wf}', workflow_name)
+    data = data.replace('{name-your-service}', os.path.split(os.getcwd())[-1])
+
+    create_file(os.path.join(workflow_path, f"config_{workflow_name}.json"), data)
+    
+    # Manipulate setup.py to include the new build type
+    add_workflow_to_setup(workflow_name)
+
+
+def workflow_exists(wf_name: str = ""):
+    """! Checks if a workflow with the given name exists
+        Args:
+                wf_name: Workflow name to be validated
+        Returns:
+            True if workflow already exists, False otherwise
+        """
+
+    return os.path.exists(os.path.join(__proj_root__, f'batch_files/workflows/{wf_name}_workflow'))
+
+
+def validate_stage(stage_name: str = "", workflow: str = ""):
+    """! Checks if a stage with the given name is valid
+    Args:
+            stage_name: Stage name to be validated
+            workflow: Workflow name to be validated
+    Returns:
+        True if stage can be created correctly, False otherwise
+    """
+
+    # Check workflow validity
+    if not workflow_exists(workflow):
+        click.echo(f'Workflow "{workflow}" does not exist. Please create or try a different workflow.')
+        return False
+
+    # Check stage validity
+    if os.path.exists(os.path.join(__proj_root__, f'batch_files/workflows/{workflow}_workflow/stages/{stage_name}')):
+        click.echo(f'Stage "{stage_name}" already exists for given workflow "{workflow}". Please use a different name.')
+        return False
+
+    return True
 
 
 def generate_documentation(simple="full"):
@@ -512,7 +724,7 @@ def generate_documentation(simple="full"):
         click.echo('Documenting project (without base).....')
 
     # Get folders
-    current_pipeline_folder = os.path.abspath(os.getcwd())
+    current_pipeline_folder = __proj_root__
     current_doc_folder = os.path.join(current_pipeline_folder, "documentation")
     if os.path.isdir(os.path.join(current_pipeline_folder, "dsframework", "documentation")):
         current_doc_folder = os.path.join(current_pipeline_folder, "dsframework", "documentation")
@@ -568,7 +780,6 @@ def get_exclude_string(venv_path):
 
 
 def get_excluded_path_list(regex_str, path, get_type):
-
     if get_type == 'folders':
         path_list = [os.path.join(path, name) for name in os.listdir(path)
                      if os.path.isdir(os.path.join(path, name))]
@@ -585,6 +796,27 @@ def get_excluded_path_list(regex_str, path, get_type):
         return path_list
 
 
+def zip_folder(local_documentation_folder, path_file_name):
+
+    if os.path.isdir(local_documentation_folder):
+        if os.listdir(local_documentation_folder):
+            shutil.make_archive(path_file_name, 'zip', local_documentation_folder)
+
+
+def hash_file(file_path):
+
+    hash_code = hashlib.sha256()
+
+    with open(file_path, 'rb') as f:
+        while True:
+            chunk = f.read(4096)
+            if not chunk:
+                break
+            hash_code.update(chunk)
+
+    return hash_code.hexdigest()
+
+
 @cli.command()
 def upload_documentation():
     """! Upload project documentation CLI
@@ -596,48 +828,40 @@ def upload_documentation():
     # Set folders
     current_working_path = os.getcwd()
     project_name = os.path.split(current_working_path)[1]
-    current_pipeline_folder = os.path.abspath(os.getcwd())
-    current_doc_folder = os.path.join(current_pipeline_folder, 'documentation', 'project_docs')
+    current_doc_folder = os.path.join(current_working_path, 'documentation', 'project_docs')
+    local_upload_folder = os.path.join(current_working_path, 'documentation', 'upload')
     html_folder_name = 'html'
 
     google_gs_path = 'gs://dozi-stg-ds-apps-1-ds-apps-ds-portal/documentation/'
-    project_gs_folder = f'{google_gs_path}{project_name}'
-    google_storage_path = 'https://storage.cloud.google.com/dozi-stg-ds-apps-1-ds-apps-ds-portal/documentation/'
-    base_tag_path = google_storage_path + project_name + '/' + html_folder_name + '/'
+
+    project_gs_folder = f'{google_gs_path}'
+
+    if project_name == 'ds-framework':
+        current_doc_folder = os.path.join(current_working_path, 'dsframework', 'documentation', 'project_docs')
+        local_upload_folder = os.path.join(current_working_path, 'dsframework', 'documentation', 'upload')
 
     # Validity checks
     if not os.path.isfile(os.path.join(current_doc_folder, html_folder_name, 'index.html')):
         click.echo('Documentation folder not found, please change directory to project root folder.')
         return
 
-    click.echo(f'Uploading {project_name} documentation started....')
+    local_documentation_folder = os.path.join(current_doc_folder, html_folder_name)
 
-    # Create backup of current docs on google storage
-    click.echo(f'Creating backup of {project_name} current documentation....')
+    if os.path.isdir(local_upload_folder):
+        shutil.rmtree(local_upload_folder)
 
-    backup_project_name = datetime.today().strftime('%Y_%m_%d_%H_%M_%S') + '_' + project_name
-    backup_destination_folder = f'{google_gs_path}backups/{backup_project_name}'
+    os.mkdir(local_upload_folder)
 
-    call_gsutil_command(f'gsutil -m mv {project_gs_folder} {backup_destination_folder}',
-                        project_name, current_doc_folder, 'Backup current documentation')
+    zip_folder(local_documentation_folder, os.path.join(local_upload_folder, project_name))
 
-    # Modify project documentation, add base tag and paths
-    click.echo(f'Adding <base> tag to {project_name} documentation....')
-    try:
-        html_list = doxygen_handler.find_files(os.path.join(current_doc_folder, html_folder_name), 'html')
-        css_list = doxygen_handler.find_files(os.path.join(current_doc_folder, html_folder_name), 'css')
-
-        doxygen_handler.modify_files(html_list, '<head>', '<head><base href=\"url_tag\">', base_tag_path)
-        doxygen_handler.modify_files(css_list, "url(\'", "url(\'url_tag", base_tag_path)
-        doxygen_handler.modify_files(css_list, "url(\"", "url(\"url_tag", base_tag_path)
-    except Exception as e:
-        click.echo(f'Adding <base> tag failed with exception: {e}')
-        return
-    else:
-        click.echo('Adding <base> tag done!')
+    hash_code = hash_file(os.path.join(local_upload_folder, project_name) + '.zip')
+    with open(os.path.join(local_upload_folder, f'{project_name}_hash.txt'), 'w') as file_hash:
+        file_hash.write(hash_code)
 
     # Upload current documentation
-    call_gsutil_command(f'gsutil -m cp -r {os.path.join(current_doc_folder, html_folder_name)} {project_gs_folder}',
+    click.echo(f'Uploading {project_name} documentation started....')
+
+    call_gsutil_command(f'gsutil -m cp -r {local_upload_folder} {project_gs_folder}',
                         project_name, current_doc_folder, 'Uploading documentation')
 
     click.echo(f'Uploading {project_name} documentation done!!!')
@@ -789,91 +1013,92 @@ def create_main_folders(target_dir, base_dir):
         os.mkdir(directories[target_dir])
 
 
-def create_pipeline_file(project_name, folder, pipeline_type, create_init_file=True):
+def create_pipeline_file(project_name, folder, pipeline_type, create_init=True):
     """! Takes the template file and replace it with new one with project_name
     @verbatim
     Args:
         project_name : The project name
         folder       : The folder we want to create the new file.
         pipeline_type : File type in the pipeline.
-        create_init_file(=True): Create init file - if doesnt exist.
+        create_init(=True): Create init file - if doesnt exist.
     @endverbatim
     """
 
     data = read_template_file(pipeline_type)
-    replace_in_template_and_create_file(project_name, folder, pipeline_type, data, create_init_file)
+    replace_in_template_and_create_file(project_name, folder, pipeline_type, data, create_init)
 
 
-def create_schema_file(project_name, folder, pipelineType, createInitFile=True):
+def create_schema_file(project_name, folder, pipeline_type, create_init=True):
     """! Create schema files (read the template and replace it )
     @verbatim
     Args:
         project_name : The project name
         folder       : The folder we want to create the new file.
-        pipelineType : File type. TODO change pipelineType name
-        createInitFile(=True): Create init file - if doesnt exist.
+        pipeline_type : File type. TODO change pipelineType name
+        create_init(=True): Create init file - if doesnt exist.
     @endverbatim
     """
-    data = read_template_file('schema/' + pipelineType)
-    replace_in_template_and_create_file(project_name, folder, pipelineType, data, createInitFile)
+    data = read_template_file('schema/' + pipeline_type)
+    replace_in_template_and_create_file(project_name, folder, pipeline_type, data, create_init)
 
 
-def create_server_file(project_name, folder, pipelineType, createInitFile=True):
+def create_server_file(project_name, folder, pipeline_type, create_init=True):
     """! Create server file (read the template and use him to create the new file)
     @verbatim
     Args:
         project_name : The project name
         folder       : The folder we want to create the new file.
-        pipelineType : File type. TODO change pipelineType name
-        createInitFile(=True): Create init file - if doesnt exist.
+        pipeline_type : File type. TODO change pipelineType name
+        create_init(=True): Create init file - if doesnt exist.
     @endverbatim
     """
-    data = read_template_file('tester/server/' + pipelineType)
-    replace_in_template_and_create_file(project_name, folder, pipelineType, data, createInitFile)
+    data = read_template_file('tester/server/' + pipeline_type)
+    replace_in_template_and_create_file(project_name, folder, pipeline_type, data, create_init)
 
 
-def create_tester_file(project_name, folder, test_file_type, createInitFile=True):
+def create_tester_file(project_name, folder, test_file_type, create_init=True):
     """! Create tester file (read the template and use him to create the new file)
     @verbatim
     Args:
         project_name   : The project name.
         folder         : The folder we want to create the new file.
         test_file_type : File type.
-        createInitFile(=True): Create init file - if doesnt exist.
+        create_init(=True): Create init file - if doesnt exist.
     @endverbatim
     """
     data = read_template_file('tester/' + test_file_type)
-    replace_in_template_and_create_file(project_name, folder, test_file_type, data, createInitFile)
+    replace_in_template_and_create_file(project_name, folder, test_file_type, data, create_init)
 
 
-def create_trainer_file(project_name, to_folder, test_file_type, createInitFile=True, from_folder=''):
+def create_trainer_file(project_name, to_folder, test_file_type, create_init=True, from_folder=''):
     """! Create trainer file (read the template and use him to create the new file)
     @verbatim
     Args:
         project_name   : The project name.
-        folder         : The folder we want to create the new file.
+        to_folder      : The folder we want to create the new file in.
         test_file_type : File type.
-        createInitFile(=True): Create init file - if it doesn't exist.
+        create_init(=True): Create init file - if it doesn't exist.
+        from_folder    : The folder from which we read the template file
     @endverbatim
     """
 
     # data = read_template_file('trainer/' + test_file_type)
     data = read_template_file(os.path.join('trainer', from_folder, test_file_type))
-    replace_in_template_and_create_file(project_name, to_folder, test_file_type, data, createInitFile)
+    replace_in_template_and_create_file(project_name, to_folder, test_file_type, data, create_init)
 
 
-def create_dsp_file(project_name, folder, test_file_type, createInitFile=True):
+def create_dsp_file(project_name, folder, test_file_type, create_init=True):
     """! Create DSP file (read the template and use him to create the new file)
     @verbatim
         Args:
             project_name   : The project name.
             folder         : The folder we want to create the new file.
             test_file_type : File type.
-            createInitFile(=True): Create init file - if doesnt exist.
+            create_init(=True): Create init file - if doesnt exist.
     @endverbatim
     """
     data = read_template_file('tester/dsp/' + test_file_type)
-    replace_in_template_and_create_file(project_name, folder, test_file_type, data, createInitFile)
+    replace_in_template_and_create_file(project_name, folder, test_file_type, data, create_init)
 
 
 def create_exist_pipeline_file(type, fileName):
@@ -908,18 +1133,20 @@ def create_exist_pipeline_file(type, fileName):
     pass
 
 
-def create_stage_files_from_template(stage_name):
-    """! Create all needed folders and files for the new stage creation
+def create_stage_files_from_template(stage_name: str = "", workflow_name: str = ""):
+    """! Create all needed folders and files for the new stage
     @verbatim
     Args:
         stage_name: Stage name
+        workflow_name: Workflow that this stage belongs to
     @endverbatim
     """
 
     data = read_template_file('batch/stage')
-    stage_directory_name = 'batch_files/stages/' + stage_name
+    stage_directory_name = os.path.join(__proj_root__,
+                                        f'batch_files/workflows/{workflow_name}_workflow/stages/{stage_name}')
     os.mkdir(stage_directory_name)
-    replace_in_stage_template_and_create_file(stage_name, stage_directory_name, data)
+    replace_in_stage_template_and_create_file(stage_name, stage_directory_name, workflow_name, data)
 
 
 def create_tester_env():
@@ -974,9 +1201,9 @@ def create_tester_env():
     pass
 
 
-def read_template_file(filename):
+def read_template_file(filename: str, file_type: str = "py"):
     """! Read \<filename\>_template.py  from the cli dir """
-    with open(os.path.join(__location__, 'dsframework/cli/' + filename + '_template.py'), 'r') as file:
+    with open(os.path.join(__location__, 'dsframework/cli/' + filename + f"_template.{file_type}"), 'r') as file:
         data = file.read()
         return data
 
@@ -1030,17 +1257,20 @@ def replace_in_template_and_create_file(project_name, folder, file_name, data, w
         create_init_file(new_init_file, new_init_export)
 
 
-def replace_in_stage_template_and_create_file(stage_name, folder, data):
+def replace_in_stage_template_and_create_file(stage_name: str, folder: str, workflow_name: str, data):
     """! Get & Replace the stage template file and create a new one with stage_name
     @verbatim
     Args:
         stage_name : The stage name
         folder     : The destination folder for the new files
+        workflow_name : Workflow name under which the stage will run
         data       : The template file
     @endverbatim
     """
 
     data = data.replace('generatedStageName', to_capitalize_no_underscore(stage_name))
+    data = data.replace('generatedWFName', workflow_name)
+
     create_file(folder + "/main.py", data)
     create_empty_init_file(folder)
 
@@ -1149,7 +1379,6 @@ def inject_to_pipeline(fileName, type, className, new_init_export):
             last_index_of_add_component = all_add_components[-1][-1]
         # finding imports and add components indexes
 
-
         index_to_add = 0
 
         # add import to end of imports or to top of file
@@ -1188,7 +1417,7 @@ def inject_to_pipeline(fileName, type, className, new_init_export):
             data = data[:first_index_of_add_component] + new_component_line + data[first_index_of_add_component:]
             data_changed = True
 
-        #adding add_component line
+        # adding add_component line
         if last_index_of_add_component > -1 and add_component_line not in data:
             last_index_of_add_component += index_to_add
             add_component_line = new_tab_line + add_component_line
@@ -1354,7 +1583,7 @@ def copy_cloud_eval_files(main_dir):
                         start_ix = data.index("#cli_file_dest=[") + len("#cli_file_dest=[")
                         end_ix = data.index("]", start_ix)
                         destination_dir = data[start_ix:end_ix]
-                        data = data[:data.index("#cli_file_dest=[")] + data[end_ix+1:]
+                        data = data[:data.index("#cli_file_dest=[")] + data[end_ix + 1:]
                         dirToCreate = main_dir + '/' + destination_dir
 
                     if not os.path.exists(dirToCreate + file_path):
@@ -1585,11 +1814,11 @@ def let_user_pick(options):
     """
     print("Please choose:")
     for idx, element in enumerate(options):
-        print("{}) {}".format(idx+1, element))
+        print("{}) {}".format(idx + 1, element))
     i = input("Enter number: ")
     try:
         if 0 < int(i) <= len(options):
-            return int(i)-1
+            return int(i) - 1
     except:
         pass
     return None
@@ -1600,8 +1829,7 @@ def deploy_service_to_gae():
     """! Run deployment scripts in order to deploy the current project to GAE.
    """
     # Setup paths to script files
-    current_folder = os.path.abspath(os.getcwd())
-    setup_file_path = f"{current_folder}/deploy/setup_deploy_params.sh"
+    setup_file_path = f"{__proj_root__}/deploy/setup_deploy_params.sh"
     framework_folder = os.path.dirname(os.path.abspath(__file__))
 
     # Check that we have a setup file. If not, copy it.
@@ -1619,9 +1847,9 @@ def deploy_service_to_gae():
 
     # Execute
     subprocess.call(full_command, shell=True)
-    #subprocess.call(setup_command, shell=True)
-    #subprocess.call(build_command, shell=True)
-    #subprocess.call(deploy_command, shell=True)
+    # subprocess.call(setup_command, shell=True)
+    # subprocess.call(build_command, shell=True)
+    # subprocess.call(deploy_command, shell=True)
 
 
 if __name__ == '__main__':
