@@ -1,9 +1,9 @@
 import subprocess
 import os
 import sys
-from json import dumps as json_dumps
-from json import load as json_load
-from json import loads as json_loads
+from ujson import dumps as json_dumps
+from ujson import load as json_load
+from ujson import loads as json_loads
 from yaml import dump as yaml_dump
 
 from datetime import timedelta
@@ -21,17 +21,23 @@ from google.cloud.dataproc_v1.types.workflow_templates import WorkflowTemplate
 # @file
 # @brief Defines workflow base class.
 class ZIDS_Workflow():
-    def __init__(self):
+    def __init__(self, workflow_name: str = ""):
         """! ZIDS_Workflow initializer.
         Loads config file and initializes parameters."""
 
-        self.trainer = False
+        self.wf_name = workflow_name
 
         file_path = sys.modules[self.__class__.__module__].__file__
         self.curr_path = file_path[:file_path.rfind("batch_files")]
 
         with open(self.curr_path + 'batch_files/batch_config.json') as batch_cfg_file:
-            self.config = json_load(batch_cfg_file)
+            batch_config = json_load(batch_cfg_file)
+
+        with open(self.curr_path + 'batch_files/workflows/' + self.wf_name +
+                  '_workflow/config_' + self.wf_name + '.json') as wf_cfg_file:
+            wf_config = json_load(wf_cfg_file)
+
+        self.config = {**batch_config, **wf_config}
 
         self.user_email = ''
         env_type = os.environ.get('SPRING_PROFILES_ACTIVE')
@@ -53,6 +59,7 @@ class ZIDS_Workflow():
         self.bucket_name = self.config['bucket_name']
         self.project_name = self.config['project_name']
         self.autoscaling_policy_id = self.project_name + '-auto-scaling-policy'
+        self.wf_stages = self.config['workflow_stages']
 
         self.unique_iteration_id = self.config['unique_iteration_id']
         self.unique_template_id = self.config['unique_template_id']
@@ -60,9 +67,10 @@ class ZIDS_Workflow():
             self.template_id = self.template_id + '_' + self.unique_template_id
 
         if self.unique_iteration_id:
-            self.folder_path = self.project_name + self.user_email + '/unique_iteration_id_' + self.unique_iteration_id
+            self.folder_path = self.project_name + self.user_email + \
+                               '/' + self.wf_name + '/unique_iteration_id_' + self.unique_iteration_id
         else:
-            self.folder_path = self.project_name + self.user_email + '/main'
+            self.folder_path = self.project_name + self.user_email + '/' + self.wf_name + '/main'
 
         self.bucket_path = f'gs://{self.bucket_name}/{self.folder_path}'
         self.project_path = 'projects/{project_id}/regions/{region}'.format(project_id=self.project_id,
@@ -80,6 +88,11 @@ class ZIDS_Workflow():
 
             cluster_init_list.append(cluster_init_action)
 
+            # Take general parameters regarding GCE configuration from the main config file, as they can be modified
+            # by Jenkins when deploying the service
+            master_gce_config = batch_config['cluster_conf']['managed_cluster']['config']['gce_cluster_config']
+            self.config['cluster_conf']['managed_cluster']['config']['gce_cluster_config'] = master_gce_config
+
         # Define workflow client
         self.serv_client = WorkflowTemplateServiceClient(
             client_options={"api_endpoint": f"{self.region}-dataproc.googleapis.com:443"}
@@ -87,14 +100,9 @@ class ZIDS_Workflow():
 
         # Define extra stage parameters
         self.extra_params = json_dumps(self.config['extra_job_params'])
-        #select if to use current date or batch_config.json as date
+        # Selection between current date or batch_config.json as date
         self.instantiate_now = False if "date_str" in self.config else True
 
-    def set_trainer_mode(self, mode: bool = False):
-        """Define trainer mode, sets the self.trainer.
-        Currently, used in the export_template_structure() to determine which stages to use."""
-        self.trainer = mode
-
     def configure_pip_packages(self, requirements_path: str, append: bool = False):
         """Configure pip packages based on the required steps.
 
@@ -122,12 +130,17 @@ class ZIDS_Workflow():
 
     def check_cluster(self):
         """! ZIDS_Workflow get_cluster.
-        Tries to locate existing cluster that should be set up by this point.
+        Tries to locate an existing cluster. If exists, we'll use it for template instantiation.
+        If not, we will use the managed cluster to create it based on configuration
+        When clusters are created manually, they include the workflow name, so this function looks for it.
+        If it was created automatically, then the name is generated and this function will not find it.
         """
         clusterClient = ClusterControllerClient(
             client_options={"api_endpoint": f"{self.region}-dataproc.googleapis.com:443"}
         )
-        cluster_name = self.config['cluster_conf']['managed_cluster']['cluster_name']
+
+        cluster_name = f"{self.config['cluster_conf']['managed_cluster']['cluster_name']}-{self.wf_name}"
+
         try:
             exist_cluster = clusterClient.get_cluster(
                 project_id=self.project_id,
@@ -148,9 +161,10 @@ class ZIDS_Workflow():
                 # If there's something other than 404, we want to know, but the flow can carry on.
                 print(f'Error getting cluster: {e}. Cluster will be created on demand')
 
-    def create_template_structure(self) -> WorkflowTemplate:
-        """! ZIDS_Workflow create_template_structure.
-        Tries to create a workflow template, containing all available stages.
+    def build_template_structure(self) -> WorkflowTemplate:
+        """! ZIDS_Workflow build_template_structure.
+        build a workflow template class instance, containing all available configurations, mainly the managed cluster
+        configuration at this point.
 
             Returns:
                 WorkflowTemplate() - Created template with basic arguments
@@ -194,7 +208,7 @@ class ZIDS_Workflow():
         pyJob.properties = spark_properties
         pyJob.args = self.default_args + additional_args
         pyJob.jar_file_uris = self.default_jars + additional_jars
-        created_stage.step_id = f'{self.project_name}-stage_{stage_name}'
+        created_stage.step_id = f'{self.project_name}-{stage_name}'
         created_stage.pyspark_job = pyJob
         return created_stage
 
@@ -202,7 +216,6 @@ class ZIDS_Workflow():
         """! Check for an existing template. If found, delete it.
         """
 
-        exist_template = None
         exist_template = self.get_workflow_template()
 
         # Deleting template if it exists
@@ -300,5 +313,5 @@ class ZIDS_Workflow():
 
         file_path = sys.modules[self.__class__.__module__].__file__
         curr_path = file_path[:file_path.rfind("workflow")]
-        with open(curr_path + '/workflow.yaml', 'w') as yml:
+        with open(curr_path + f'/workflow_{self.wf_name}.yaml', 'w') as yml:
             yaml_dump(workflow_json, yml)
