@@ -1,7 +1,9 @@
 import os
 from google.protobuf.duration_pb2 import Duration
 from google.cloud.dataproc_v1 import ClusterControllerClient, Cluster, ClusterConfig
+from google.cloud.dataproc_v1 import JobControllerClient, Job, PySparkJob
 from google.cloud import storage
+from ujson import dumps as ujson_dumps
 import subprocess
 import sys
 
@@ -35,6 +37,7 @@ class ZIDSBatchHelper:
             user_email = subprocess.check_output(command, shell=True).decode(sys.stdout.encoding)
             self.user_email = '/' + user_email.split("@")[0]
 
+        self.wf_name = stage_config['workflow_name']
         self.bucket_name = stage_config['bucket_name']
         self.project_id = stage_config['project_id']
         self.project_name = stage_config['project_name']
@@ -42,15 +45,15 @@ class ZIDSBatchHelper:
         self.zone = self.region + '-b'  # TODO: Check if we can configure this as well
         self.template_id = stage_config['template_id']
         self.unique_iteration_id = stage_config['unique_iteration_id']
-        self.cluster_name = stage_config['cluster_conf']['managed_cluster']['cluster_name']
+        self.cluster_name = f"{stage_config['cluster_conf']['managed_cluster']['cluster_name']}-{self.wf_name}"
         self.cluster_duration = stage_config['cluster_duration_sec']
         self.managed_cluster = stage_config['cluster_conf']['managed_cluster']
 
         if self.unique_iteration_id:
             self.folder_path = self.project_name + self.user_email + \
-                               '/unique_iteration_id_' + self.unique_iteration_id
+                               '/' + self.wf_name + '/unique_iteration_id_' + self.unique_iteration_id
         else:
-            self.folder_path = self.project_name + self.user_email + '/main'
+            self.folder_path = self.project_name + self.user_email + '/' + self.wf_name + '/main'
 
         self.bucket_path = f'gs://{self.bucket_name}/{self.folder_path}'
         self.project_path = 'projects/{project_id}/regions/{region}'.format(project_id=self.project_id,
@@ -74,7 +77,13 @@ class ZIDSBatchHelper:
         with open(main_project_path + '/requirements_docker.txt') as f:
             required = f.read().splitlines()
             required = [x for x in required if "#" not in x]
-            pip_packages = ' '.join(required)
+
+        with open(main_project_path +
+                  f"/batch_files/workflows/{self.wf_name}_workflow/requirements_{self.wf_name}.txt") as f:
+            wf_required = f.read().splitlines()
+            wf_required = [x for x in wf_required if "#" not in x]
+
+        pip_packages = ' '.join(required + wf_required)
 
         client = ClusterControllerClient(
             client_options={"api_endpoint": f"{self.region}-dataproc.googleapis.com:443"}
@@ -289,6 +298,65 @@ class ZIDSBatchHelper:
             except subprocess.CalledProcessError as e:
                 print(f'Failed on connection to cluster {self.cluster_name}, Error: {e}')
 
+    def run_single_stage(self, stage_name: str):
+        """! Run a specific batch stage on an existing cluster
+            Args:
+                stage_name: Stage name to execute as a job
+        """
+
+        # Define some job parameters
+        stage_dir = f"batch_files/workflows/{self.wf_name}_workflow/stages/{stage_name}"
+        file_name = f"{self.project_name}_{self.wf_name}.zip"
+
+        spark_properties = self.config["spark_properties"]
+        if stage_name in spark_properties:
+            stage_spark_properties = {**spark_properties['basic'], **spark_properties[stage_name]}
+        else:
+            stage_spark_properties = spark_properties['basic']
+
+        spark_properties_list = [[key, value] for key, value in list(stage_spark_properties.items())]
+
+        default_jars = ['gs://spark-lib/bigquery/spark-bigquery-latest_2.12.jar']
+        default_args = [ujson_dumps(self.config),
+                        "start_date_placeholder",
+                        "end_date_placeholder",
+                        "params_placeholder"]
+
+        _, existing_cluster = self.get_client_check_cluster()
+
+        if existing_cluster:
+            try:
+                # Create a job client
+                client = JobControllerClient(
+                    client_options={"api_endpoint": f"{self.region}-dataproc.googleapis.com:443"})
+
+                # Create job to run
+                created_job = Job()
+                py_job = PySparkJob()
+
+                py_job.main_python_file_uri = f'{self.bucket_path}/{stage_dir}/main.py'
+                py_job.archive_uris = [f'{self.bucket_path}/dist/{file_name}']
+                py_job.python_file_uris = [f'{self.bucket_path}/dist/{file_name}']
+                py_job.properties = spark_properties_list
+                py_job.args = default_args
+                py_job.jar_file_uris = default_jars
+
+                created_job.pyspark_job = py_job
+                created_job.placement.cluster_name = self.cluster_name
+
+                # Submit job
+                client.submit_job(
+                    project_id=self.project_id,
+                    region=self.region,
+                    job=created_job
+                )
+                print(f'Job submitted: {stage_name}.')
+
+            except Exception as e:
+                print(f'Failed submitting job {stage_name} to cluster: {self.cluster_name}')
+        else:
+            print(f"No cluster {self.cluster_name} found. Please create and rerun command")
+
     def upload_files_to_bucket(self, upload_type: str):
         """! Upload jars, stages, cluster init script and wheel file to a given bucket
                 Args:
@@ -339,14 +407,15 @@ class ZIDSBatchHelper:
 
         for path, subdirs, files in os.walk(root_dir + '/jars'):
             for name in files:
-                path_local = os.path.join(path, name)
-                bucket_blob_path = path_local[path_local.index(self.project_name) +
-                                              len(self.project_name):].replace('\\', '/')
-                blob = self.bucket.blob(self.folder_path + bucket_blob_path)
-                blob.upload_from_filename(path_local)
+                if name.endswith('.jar'):
+                    path_local = os.path.join(path, name)
+                    bucket_blob_path = path_local[path_local.index(self.project_name) +
+                                                  len(self.project_name):].replace('\\', '/')
+                    blob = self.bucket.blob(self.folder_path + bucket_blob_path)
+                    blob.upload_from_filename(path_local)
             # print(path_local)
             # print(bucket_blob_path)
-        print('uploaded jars main successfully')
+        print('uploaded jars successfully')
 
     def upload_stages_main_to_bucket(self, root_dir):
         """! Upload relevant stages to a given bucket
@@ -354,12 +423,12 @@ class ZIDSBatchHelper:
                 root_dir: Directory in which we search for the stages directory
         """
 
-        for path, subdirs, files in os.walk(root_dir + '/stages'):
+        for path, subdirs, files in os.walk(root_dir + f"/workflows/{self.wf_name}_workflow/stages"):
             for name in files:
                 if name == 'main.py':
                     path_local = os.path.join(path, name)
-                    bucket_blob_path = path_local[path_local.index(self.project_name) +
-                                                  len(self.project_name):].replace('\\', '/')
+                    stage_name = path_local.rsplit('/', 2)[-2]  # Stage name will always be the parent dir of main.py
+                    bucket_blob_path = f"/batch_files/stages/{stage_name}/{name}"
                     blob = self.bucket.blob(self.folder_path + bucket_blob_path)
                     blob.upload_from_filename(path_local)
                 # print(path_local)
@@ -372,16 +441,39 @@ class ZIDSBatchHelper:
             Args:
                 root_dir: Directory in which we search for the wheel archive
         """
+        from concurrent.futures import ThreadPoolExecutor
+        import functools
 
         parent_path = os.path.abspath(os.path.join(root_dir, ".."))
         dist = parent_path + '/dist'
-        onlyfiles = [os.path.join(dist, f) for f in os.listdir(dist) if os.path.isfile(os.path.join(dist, f))]
-        for path_local in onlyfiles:
-            bucket_blob_path = path_local[path_local.index(self.project_name) +
-                                          len(self.project_name):].replace('\\', '/')
-            blob = self.bucket.blob(self.folder_path + bucket_blob_path)
+        onlyfiles = [os.path.join(dist, f) for f in os.listdir(dist) if os.path.isfile(os.path.join(dist, f))
+                     and self.wf_name in f]
+
+        def copy_file(path_local, project_name, folder_path, dest_bucket):
+            bucket_blob_path = path_local[path_local.index(project_name) +
+                                          len(project_name):].replace('\\', '/')
+            blob = dest_bucket.blob(folder_path + bucket_blob_path)
             blob.chunk_size = 1024 * 1024
             blob.upload_from_filename(path_local)
+            return
+
+        # Parallel run. Keep previous method as fallback.
+        if True:
+            copy_file_part = functools.partial(copy_file, project_name=self.project_name,
+                                               folder_path=self.folder_path, dest_bucket=self.bucket)
+            with ThreadPoolExecutor(max_workers=5) as executor:
+                for filename in onlyfiles:
+                    executor.submit(copy_file_part, filename)
+
+        else:
+            for path_local in onlyfiles:
+                if self.wf_name not in path_local:
+                    continue
+                bucket_blob_path = path_local[path_local.index(self.project_name) +
+                                              len(self.project_name):].replace('\\', '/')
+                blob = self.bucket.blob(self.folder_path + bucket_blob_path)
+                blob.chunk_size = 1024 * 1024
+                blob.upload_from_filename(path_local)
         # print(path_local)
         # print(project_bucket_folder + bucket_blob_path)
 
