@@ -1,4 +1,5 @@
 """! @brief Skeleton for implementing a specific stage to be used in a DataProc Workflow template."""
+import argparse
 import sys
 import logging
 from json import loads as json_loads
@@ -6,6 +7,7 @@ from dsframework.base.batch.pipeline_stage_base import ZIDS_BatchPipelineStage
 
 logger = logging.getLogger(__name__)
 
+
 ##
 # @file
 # @brief generatedStageName class, inherits ZIDS_BatchPipelineStage base class.
@@ -23,6 +25,7 @@ class generatedStageName(ZIDS_BatchPipelineStage):
             Args:
                 stage_config : Configuration dictionary, loaded from configuration file.
         """
+        self.wf_name = "generatedWFName"
         super().__init__(stage_config, logger)
 
     def load(self):
@@ -50,18 +53,26 @@ if __name__ == "__main__":
             System argument 3 - End date, received from Airflow
             System argument 4 - Placeholder for future parameters
     """
-    if sys.argv and len(sys.argv) > 1:
-        config = json_loads(sys.argv[1])
-        stage = generatedStageName(config)
-        try:
-            start_date = sys.argv[2]
-            end_date = sys.argv[3]
-            params = sys.argv[4]
-            stage.update_stage_params(start_date, end_date, params)
-            stage()
-        except Exception as e:
-            raise Exception(f" Stage failed with error: {e}")
-    else:
-        raise Exception(f"Stage configuration not provided, Can't run stage")
+
+    # Handle input arguments
+    parser = argparse.ArgumentParser(description='Configure and run generatedStageName class')
+    parser.add_argument('config', type=str,
+                        help='Configuration file')
+    parser.add_argument('start_date', type=str, default='',
+                        help='Start date, received from Airflow')
+    parser.add_argument('end_date', type=str, default='',
+                        help='End date, received from Airflow')
+    parser.add_argument('params', nargs=argparse.REMAINDER,
+                        help='Placeholder for future parameters')
+    args = parser.parse_args()
+
+    # Configure stage class and run
+    stage = generatedStageName(json_loads(args.config))
+    try:
+        stage.update_stage_params(args.start_date, args.end_date, args.params)
+        stage()
+    except Exception as e:
+        raise Exception(f" Stage failed with error: {e}")
+
 
 
