@@ -14,6 +14,19 @@ class tokenGenerator():
 	# account example - original_project_name@dozi-stg-ds-apps-1.iam.gserviceaccount.com
 	# audience example - 1042519062770-rghbs1ied8kno7l00mlnko5r7sgjea2o.apps.googleusercontent.com
 	conf = {
+		"STG_DSP_Conf": {
+			"account": 'jwt-ds-portal@dozi-stg-ds-apps-1.iam.gserviceaccount.com',
+			"host": 'dozi-stg-ds-apps-1',
+			"target_audience": '1042519062770-rghbs1ied8kno7l00mlnko5r7sgjea2o.apps.googleusercontent.com',
+			"iap": True,
+		},
+		"PRD_DSP_Conf": {
+			"account": 'jwt-ds-portal@dozi-prd-ds-apps-1.iam.gserviceaccount.com',
+			"host": 'dozi-prd-ds-apps-1',
+			"target_audience": '484284403458-lpmc6edp4b6rd2383dgpkiq5a9ngkpbc.apps.googleusercontent.com',
+			# need to get from dev ops
+			"iap": True,
+		},
 		"STG_Conf": {
 			"account": '', # need to get from dev ops
 			"host": 'dozi-stg-ds-apps-1',
@@ -29,7 +42,9 @@ class tokenGenerator():
 	}
 
 	def __init__(self):
-		tokens = {
+		self.tokens = {
+			"STG_DSP_Conf": '',
+			"PRD_DSP_Conf": '',
 			"STG_Conf": '',
 			"PRD_Conf": '',
 		}
